from elroy.config import ElroyContext
from elroy.store.data_models import UserPreference
from elroy.system.constants import UNKNOWN


def set_user_preferred_name(context: ElroyContext, preferred_name: str) -> None:
    """
    Set the user's preferred name.

    Args:
        user_id: The user's ID.
        preferred_name: The user's preferred name.
    """

    user_preference = _get_user_preference(context)

    user_preference.preferred_name = preferred_name
    context.session.commit()


def get_user_preferred_name(context: ElroyContext) -> str:
    """Returns the user's preferred name.

    Args:
        user_id (int): the user ID

    Returns:
        str: String representing the user's preferred name.
    """

    user_preference = _get_user_preference(context)

    return user_preference.preferred_name or UNKNOWN


def set_user_full_name(context: ElroyContext, full_name: str) -> str:
    """Sets the user's full name.

    Args:
        user_id (int): user id
        full_name (str): The full name of the user

    Returns:
        str: result of the attempt to set the user's full name
    """

    user_preference = _get_user_preference(context)

    old_full_name = user_preference.full_name or UNKNOWN
    user_preference.full_name = full_name
    context.session.commit()

    return f"Full name set to {full_name}. Previous value was {old_full_name}."


def get_user_full_name(context: ElroyContext) -> str:
    """Returns the user's full name.

    Args:
        user_id (int): the user ID

    Returns:
        str: String representing the user's full name.
    """

    user_preference = _get_user_preference(context)

    return user_preference.full_name or "Unknown name"


def _get_user_preference(context: ElroyContext):
    from sqlmodel import select

    user_preference = context.session.exec(
        select(UserPreference).where(
            UserPreference.user_id == context.user_id,
            UserPreference.is_active == True,
        )
    ).first()

    if user_preference is None:
        user_preference = UserPreference(user_id=context.user_id, is_active=True)
        context.session.add(user_preference)
        context.session.commit()
        context.session.refresh(user_preference)
    return user_preference



from contextlib import contextmanager
from typing import Generator, Iterator, List, Set, Text, Union

from prompt_toolkit import HTML, PromptSession
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import Style
from pygments.lexers.special import TextLexer
from rich.console import Console
from rich.panel import Panel
from rich.pretty import Pretty
from rich.text import Text

from elroy.io.base import ElroyIO
from elroy.store.data_models import FunctionCall


class CliIO(ElroyIO):
    def __init__(
        self,
        system_message_color: str,
        assistant_message_color: str,
        user_input_color: str,
        warning_color: str,
    ) -> None:
        self.console = Console()
        self.system_message_color = system_message_color
        self.assistant_message_color = assistant_message_color
        self.warning_color = warning_color
        self.user_input_color = user_input_color
        self.style = Style.from_dict(
            {
                "prompt": "bold",
                "user-input": self.user_input_color + " bold",
                "": self.user_input_color,
                "pygments.literal.string": f"bold italic {self.user_input_color}",
            }
        )

        self.prompt_session = PromptSession(
            history=InMemoryHistory(),
            style=self.style,
            lexer=PygmentsLexer(TextLexer),
        )

    def assistant_msg(self, message: Union[str, Pretty, Iterator[str], Generator[str, None, None]]) -> None:
        if isinstance(message, (Iterator, Generator)):
            try:
                for chunk in message:
                    self.console.print(f"[{self.assistant_message_color}]{chunk}[/]", end="")
            except KeyboardInterrupt:
                self.console.print()
                return

        elif isinstance(message, Pretty):
            self.console.print(message)
        else:
            self.console.print(f"[{self.assistant_message_color}]{message}[/]", end="")
        self.console.print()  # New line after complete response

    def sys_message(self, message: Union[str, Pretty]) -> None:
        if isinstance(message, Pretty):
            self.console.print(message)
        else:
            self.console.print(f"[{self.system_message_color}]{message}[/]", end="")
        self.console.print()  # New line after complete response

    def notify_function_call(self, function_call: FunctionCall) -> None:
        msg = f"[{self.system_message_color}]Executing function call: [bold]{function_call.function_name}[/bold]"
        if function_call.arguments:
            self.console.print(msg + f" with arguments:[/]", Pretty(function_call.arguments))
        else:
            self.console.print(msg + "[/]")
        self.console.print()

    def notify_function_call_error(self, function_call: FunctionCall, error: Exception) -> None:
        self.console.print(f"[{self.system_message_color}]Error executing function call: [bold]{function_call.function_name}[/bold][/]")
        self.console.print(f"[{self.system_message_color}]{error}[/]")
        self.console.print()

    def notify_warning(self, message: str) -> None:
        self.console.print(Text(message, justify="center", style=self.warning_color))  # type: ignore
        self.console.print(f"[{self.warning_color}]Please provide feedback at https://github.com/elroy-bot/elroy/issues[/]")
        self.console.print()

    def print_memory_panel(self, titles: List[str]):
        if titles:
            panel = Panel("\n".join(titles), title="Relevant Context", expand=False, border_style=self.user_input_color)
            self.console.print(panel)

    @contextmanager
    def status(self, message: str) -> Generator[None, None, None]:
        self.console.print(f"[{self.system_message_color}]{message}[/]")
        yield
        self.console.print(f"[{self.system_message_color}]Done![/]")

    def print_title_ruler(self):
        self.console.rule(
            Text("Elroy", justify="center", style=self.user_input_color),
            style=self.user_input_color,
        )

    def rule(self):
        self.console.rule(style=self.user_input_color)

    async def prompt_user(self, prompt=">") -> str:
        return await self.prompt_session.prompt_async(HTML(f"<b>{prompt} </b>"), style=self.style)

    def update_completer(self, goal_names: Set[str], memory_names: Set[str]) -> None:
        self.prompt_session.completer = SlashCompleter(goal_names, memory_names)


class SlashCompleter(WordCompleter):
    def __init__(self, goals: Set[str], memories: Set[str]):
        self.goals = goals
        self.memories = memories
        super().__init__(self.get_words(), sentence=True, pattern=r"^/")  # type: ignore

    def get_words(self):
        from elroy.tools.system_commands import (GOAL_COMMANDS,
                                                 MEMORY_COMMANDS,
                                                 SYSTEM_COMMANDS)

        words = [f"/{f.__name__}" for f in SYSTEM_COMMANDS - (GOAL_COMMANDS | MEMORY_COMMANDS)]
        words += [f"/{f.__name__} {goal}" for f in GOAL_COMMANDS for goal in self.goals]
        words += [f"/{f.__name__} {memory}" for f in MEMORY_COMMANDS for memory in self.memories]
        return words



from dataclasses import asdict
from datetime import timedelta
from typing import Dict, Iterable, List, Optional

from sqlmodel import select
from toolz import first, pipe
from toolz.curried import filter, map, pipe

from elroy.config import ElroyContext
from elroy.store.data_models import (USER, ContextMessage, ContextMessageSet,
                                     MemoryMetadata, Message)
from elroy.system.clock import ensure_utc, get_utc_now
from elroy.system.parameters import CHAT_MODEL
from elroy.system.utils import last_or_none, logged_exec_time


def context_message_to_db_message(user_id: int, context_message: ContextMessage):

    return Message(
        id=context_message.id,
        user_id=user_id,
        content=context_message.content,
        role=context_message.role,
        model=CHAT_MODEL,
        tool_calls=[asdict(t) for t in context_message.tool_calls] if context_message.tool_calls else None,
        tool_call_id=context_message.tool_call_id,
        memory_metadata=[asdict(m) for m in context_message.memory_metadata],
    )


def db_message_to_context_message_dict(db_message: Message) -> Dict:
    return {
        "id": db_message.id,
        "content": db_message.content,
        "role": db_message.role,
        "created_at": ensure_utc(db_message.created_at),
        "tool_calls": db_message.tool_calls,
        "tool_call_id": db_message.tool_call_id,
        "memory_metadata": [MemoryMetadata(**m) for m in db_message.memory_metadata] if db_message.memory_metadata else [],
    }


def get_current_context_message_set_db(context: ElroyContext) -> Optional[ContextMessageSet]:
    return context.session.exec(
        select(ContextMessageSet).where(
            ContextMessageSet.user_id == context.user_id,
            ContextMessageSet.is_active == True,
        )
    ).first()


def get_time_since_context_message_creation(context: ElroyContext) -> Optional[timedelta]:
    row = get_current_context_message_set_db(context)

    if row:
        return get_utc_now() - ensure_utc(row.created_at)


def _get_context_messages_iter(context: ElroyContext) -> Iterable[ContextMessage]:
    # TODO: Cache this
    def get_message_dict(id: int) -> Dict:
        msg = context.session.exec(select(Message).where(Message.id == id)).first()
        assert msg
        return db_message_to_context_message_dict(msg)

    agent_context = get_current_context_message_set_db(context)

    return pipe(
        [] if not agent_context else agent_context.message_ids,
        map(get_message_dict),
        map(lambda d: ContextMessage(**d)),
        list,
    )  # type: ignore


def get_current_system_message(context: ElroyContext) -> Optional[ContextMessage]:
    try:
        return first(_get_context_messages_iter(context))
    except StopIteration:
        return None


@logged_exec_time
def get_time_since_most_recent_user_message(context: ElroyContext) -> Optional[timedelta]:
    return pipe(
        _get_context_messages_iter(context),
        filter(lambda x: x.role == USER),
        last_or_none,
        lambda x: get_utc_now() - x.created_at if x else None,
    )  # type: ignore


@logged_exec_time
def get_context_messages(context: ElroyContext) -> List[ContextMessage]:
    return list(_get_context_messages_iter(context))


def persist_messages(context: ElroyContext, messages: List[ContextMessage]) -> List[int]:
    msg_ids = []
    for msg in messages:
        if msg.id:
            msg_ids.append(msg.id)
        else:
            db_message = context_message_to_db_message(context.user_id, msg)
            context.session.add(db_message)
            context.session.commit()
            context.session.refresh(db_message)
            msg_ids.append(db_message.id)
    return msg_ids


def remove_context_messages(context: ElroyContext, messages: List[ContextMessage]) -> None:
    assert all(m.id is not None for m in messages), "All messages must have an id to be removed"

    msg_ids = [m.id for m in messages]

    replace_context_messages(context, [m for m in get_context_messages(context) if m.id not in msg_ids])


def add_context_messages(context: ElroyContext, messages: List[ContextMessage]) -> None:
    replace_context_messages(
        context,
        get_context_messages(context) + messages,
    )


def replace_context_messages(context: ElroyContext, messages: List[ContextMessage]) -> None:
    msg_ids = persist_messages(context, messages)

    existing_context = context.session.exec(
        select(ContextMessageSet).where(
            ContextMessageSet.user_id == context.user_id,
            ContextMessageSet.is_active == True,
        )
    ).first()

    if existing_context:
        existing_context.is_active = None
        context.session.add(existing_context)
    new_context = ContextMessageSet(
        user_id=context.user_id,
        message_ids=msg_ids,
        is_active=True,
    )
    context.session.add(new_context)
    context.session.commit()



import asyncio
import logging
import os
import sys
import threading
from contextlib import contextmanager
from datetime import datetime, timedelta
from typing import Generator, Optional

import typer
from colorama import init
from toolz import pipe

from elroy.cli.updater import (check_updates, ensure_current_db_migration,
                               version_callback)
from elroy.config import ROOT_DIR, ElroyContext, get_config, session_manager
from elroy.docker_postgres import is_docker_running, start_db, stop_db
from elroy.io.base import ElroyIO, StdIO
from elroy.io.cli import CliIO
from elroy.logging_config import setup_logging
from elroy.memory import get_memory_names, get_relevant_memories
from elroy.onboard_user import onboard_user
from elroy.store.data_models import SYSTEM, USER
from elroy.store.goals import get_goal_names
from elroy.store.message import get_time_since_most_recent_user_message
from elroy.store.user import is_user_exists
from elroy.system.parameters import (CLI_USER_ID, DEFAULT_ASSISTANT_COLOR,
                                     DEFAULT_INPUT_COLOR,
                                     DEFAULT_SYSTEM_MESSAGE_COLOR,
                                     DEFAULT_WARNING_COLOR,
                                     MIN_CONVO_AGE_FOR_GREETING)
from elroy.system.utils import datetime_to_string
from elroy.system_context import context_refresh
from elroy.tools.functions.user_preferences import (get_user_preferred_name,
                                                    set_user_preferred_name)
from elroy.tools.messenger import process_message
from elroy.tools.system_commands import (SYSTEM_COMMANDS, contemplate,
                                         invoke_system_command)

app = typer.Typer(help="Elroy CLI", context_settings={"obj": {}})


@contextmanager
def init_elroy_context(ctx: typer.Context, io: Optional[ElroyIO] = None) -> Generator[ElroyContext, None, None]:
    """Create an ElroyContext as a context manager"""

    if not io:
        io = CliIO(
            ctx.obj["system_message_color"],
            ctx.obj["assistant_color"],
            ctx.obj["user_input_color"],
            ctx.obj["warning_color"],
        )

    try:
        setup_logging(ctx.obj["log_file_path"])

        if ctx.obj["use_docker_postgres"]:
            if ctx.obj["postgres_url"] is not None:
                logging.info("postgres_url is set, ignoring use_docker_postgres set to True")

            else:
                if not is_docker_running():
                    io.sys_message(
                        "Docker is not running, and elroy started with --use_docker_postgres. Please start Docker and try again."
                    )
                    exit(1)
                ctx.obj["postgres_url"] = start_db()

        assert ctx.obj["postgres_url"], "Database URL is required"
        assert ctx.obj["openai_api_key"], "OpenAI API key is required"

        # Check if migrations need to be run
        ensure_current_db_migration(io, ctx.obj["postgres_url"])

        config = get_config(
            postgres_url=ctx.obj["postgres_url"],
            openai_api_key=ctx.obj["openai_api_key"],
            context_window_token_limit=ctx.obj["context_window_token_limit"],
        )

        with session_manager(config.postgres_url) as session:
            yield ElroyContext(
                user_id=CLI_USER_ID,
                session=session,
                config=config,
                io=io,
            )

    finally:
        if ctx.obj["use_docker_postgres"] and ctx.obj["stop_docker_postgres_on_exit"]:
            io.sys_message("Stopping Docker Postgres container...")
            stop_db()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(None, "--version", callback=version_callback, is_eager=True, help="Show version and exit."),
    postgres_url: Optional[str] = typer.Option(
        None, envvar="ELROY_POSTGRES_URL", help="Postgres URL to use for Elroy. If set, ovverrides use_docker_postgres."
    ),
    openai_api_key: Optional[str] = typer.Option(None, envvar="OPENAI_API_KEY", help="OpenAI API key, required."),
    context_window_token_limit: Optional[int] = typer.Option(
        None, envvar="ELROY_CONTEXT_WINDOW_TOKEN_LIMIT", help="How many tokens to keep in context before compressing."
    ),
    log_file_path: str = typer.Option(
        os.path.join(ROOT_DIR, "logs", "elroy.log"), envvar="ELROY_LOG_FILE_PATH", help="Where to write logs."
    ),
    use_docker_postgres: Optional[bool] = typer.Option(
        True,
        envvar="USE_DOCKER_POSTGRES",
        help="If true and postgres_url is not set, will attempt to start a Docker container for Postgres.",
    ),
    stop_docker_postgres_on_exit: Optional[bool] = typer.Option(
        False, envvar="STOP_DOCKER_POSTGRES_ON_EXIT", help="Whether or not to stop the Postgres container on exit."
    ),
    system_message_color: str = typer.Option(DEFAULT_SYSTEM_MESSAGE_COLOR, help="Color for system messages."),
    user_input_color: str = typer.Option(DEFAULT_INPUT_COLOR, help="Color for user input."),
    assistant_color: str = typer.Option(DEFAULT_ASSISTANT_COLOR, help="Color for assistant output."),
    warning_color: str = typer.Option(DEFAULT_WARNING_COLOR, help="Color for warning messages."),
):
    """Common parameters."""
    ctx.obj = {
        "postgres_url": postgres_url,
        "openai_api_key": openai_api_key,
        "context_window_token_limit": context_window_token_limit,
        "log_file_path": log_file_path,
        "use_docker_postgres": use_docker_postgres,
        "stop_docker_postgres_on_exit": stop_docker_postgres_on_exit,
        "system_message_color": system_message_color,
        "user_input_color": user_input_color,
        "assistant_color": assistant_color,
        "warning_color": warning_color,
    }


@app.command()
def chat(ctx: typer.Context):
    """Start the Elroy chat interface"""

    if not sys.stdin.isatty():
        with init_elroy_context(ctx, StdIO()) as context:
            for line in sys.stdin:
                process_and_deliver_msg(context, line)
        return

    with init_elroy_context(ctx) as context:
        check_updates(context)
        asyncio.run(main_chat(context))
        context.io.sys_message(f"Exiting...")


def process_and_deliver_msg(context: ElroyContext, user_input: str, role=USER):
    if user_input.startswith("/") and role == USER:
        cmd = user_input[1:].split()[0]

        if cmd.lower() not in {f.__name__ for f in SYSTEM_COMMANDS}:
            context.io.assistant_msg(f"Unknown command: {cmd}")
        else:
            try:
                context.io.sys_message(invoke_system_command(context, user_input))
            except Exception as e:
                context.io.sys_message(f"Error invoking system command: {e}")
    else:
        context.io.assistant_msg(process_message(context, user_input, role))

    context.io.rule()


def periodic_context_refresh(context: ElroyContext, interval_seconds: float):
    """Run context refresh in a background thread"""
    # Create new event loop for this thread
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    async def refresh_loop(context: ElroyContext):
        while True:
            try:
                logging.info("Refreshing context")
                await context_refresh(context)  # Keep this async
                await asyncio.sleep(interval_seconds)
            except Exception as e:
                logging.error(f"Error in periodic context refresh: {e}")
                context.session.rollback()

    try:
        # hack to get a new session for the thread
        with session_manager(context.config.postgres_url) as session:
            loop.run_until_complete(
                refresh_loop(
                    ElroyContext(
                        user_id=CLI_USER_ID,
                        session=session,
                        config=context.config,
                        io=context.io,
                    )
                )
            )
    finally:
        loop.close()


def run_in_background_thread(fn, context, *args):
    # hack to get a new session for the thread
    with session_manager(context.config.postgres_url) as session:
        thread = threading.Thread(
            target=fn,
            args=(ElroyContext(user_id=CLI_USER_ID, session=session, config=context.config, io=context.io), *args),
            daemon=True,
        )
        thread.start()


async def main_chat(context: ElroyContext[CliIO]):
    init(autoreset=True)

    run_in_background_thread(
        periodic_context_refresh,
        context,
        context.config.context_refresh_interval_seconds,
    )

    context.io.print_title_ruler()

    if not is_user_exists(context):
        context.io.notify_warning("Elroy is in alpha release")
        name = await context.io.prompt_user("Welcome to Elroy! What should I call you?")
        user_id = onboard_user(context.session, context.io, context.config, name)
        assert isinstance(user_id, int)

        set_user_preferred_name(context, name)
        process_and_deliver_msg(context, "Elroy user {name} has been onboarded. Say hello and introduce yourself.", role=SYSTEM)

    elif (get_time_since_most_recent_user_message(context) or timedelta()) < MIN_CONVO_AGE_FOR_GREETING:
        logging.info("User has interacted recently, skipping greeting.")
    else:
        preferred_name = get_user_preferred_name(context)

        process_and_deliver_msg(
            context,
            f"{preferred_name} has logged in. The current time is {datetime_to_string(datetime.now())}. I should offer a brief greeting.",
            SYSTEM,
        )

    while True:
        try:

            context.io.update_completer(get_goal_names(context), get_memory_names(context))
            pipe(context, get_relevant_memories, context.io.print_memory_panel)

            user_input = await get_user_input(context)
            if user_input.lower().startswith("/exit") or user_input == "exit":
                break
            elif user_input:
                process_and_deliver_msg(context, user_input)
                run_in_background_thread(contemplate, context)
        except EOFError:
            break


async def get_user_input(context: ElroyContext, keyboard_interupt_count=0) -> str:
    try:
        return await context.io.prompt_user()
    except KeyboardInterrupt:
        keyboard_interupt_count += 1
        if keyboard_interupt_count == 3:
            context.io.assistant_msg("To exit, type /exit, exit, or Ctrl-D")
        if keyboard_interupt_count > 5:
            raise EOFError
        else:
            return await get_user_input(context, keyboard_interupt_count)


def main():
    if len(sys.argv) == 1:
        sys.argv.append("chat")
    app()


if __name__ == "__main__":
    main()



import logging
from functools import wraps
from typing import Callable, TypeVar

from elroy.io.cli import CliIO

T = TypeVar("T")


def experimental(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        context = next((arg for arg in args if hasattr(arg, "io")), None)
        if not context:
            context = next((value for value in kwargs.values() if hasattr(value, "io")), None)

        if context and hasattr(context, "io"):
            io = context.io
            assert isinstance(io, CliIO)
            io.notify_warning("Warning: This is an experimental feature.")
        else:
            logging.warning("No context found to notify of experimental feature.")
        return func(*args, **kwargs)

    return wrapper


def debug(value: T) -> T:
    import pdb
    import traceback

    for line in traceback.format_stack():
        print(line.strip())
    pdb.set_trace()
    return value


def debug_log(value: T) -> T:
    import traceback

    traceback.print_stack()
    print(f"CURRENT VALUE: {value}")
    return value



import logging
from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import datetime
from typing import Generator, Iterator, Union

from elroy.store.data_models import FunctionCall


class ElroyIO(ABC):

    @abstractmethod
    def sys_message(self, message: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def assistant_msg(self, message: Union[str, Iterator[str], Generator[str, None, None]]) -> None:
        raise NotImplementedError

    @abstractmethod
    def notify_function_call(self, function_call: FunctionCall) -> None:
        raise NotImplementedError

    @abstractmethod
    def notify_function_call_error(self, function_call: FunctionCall, error: Exception) -> None:
        raise NotImplementedError

    @abstractmethod
    def notify_warning(self, message: str) -> None:
        raise NotImplementedError

    @contextmanager
    def status(self, message: str) -> Generator[None, None, None]:
        logging.info(message)
        yield

    def rule(self) -> None:
        pass


class StdIO(ElroyIO):
    """
    IO which emits plain text to stdin and stdout.
    """

    def sys_message(self, message: str) -> None:
        logging.info(f"[{datetime.now()}] SYSTEM: {message}")

    def assistant_msg(self, message: Union[str, Iterator[str], Generator[str, None, None]]) -> None:
        if isinstance(message, (Iterator, Generator)):
            message = "".join(message)
        print(message)

    def notify_function_call(self, function_call: FunctionCall) -> None:
        logging.info(f"[{datetime.now()}] FUNCTION CALL: {function_call.function_name}({function_call.arguments})")

    def notify_function_call_error(self, function_call: FunctionCall, error: Exception) -> None:
        logging.error(f"[{datetime.now()}] FUNCTION ERROR: {function_call.function_name} - {str(error)}")

    def notify_warning(self, message: str) -> None:
        logging.warning(message)


class TestIO(ElroyIO):
    """
    IO which simply logs calls to its methods, with output.
    """

    def __init__(self):
        self.system_messages: list[str] = []
        self.assistant_messages: list[str] = []
        self.function_calls: list[FunctionCall] = []
        self.function_call_errors: list[tuple[FunctionCall, Exception]] = []
        self.warnings: list[str] = []

    def sys_message(self, message: str) -> None:
        print(f"[{datetime.now()}] SYSTEM: {message}")
        self.system_messages.append(message)

    def assistant_msg(self, message: Union[str, Iterator[str], Generator[str, None, None]]) -> None:
        if isinstance(message, (Iterator, Generator)):
            message = "".join(message)
        print(f"[{datetime.now()}] ASSISTANT: {message}")
        self.assistant_messages.append(message)

    def notify_function_call(self, function_call: FunctionCall) -> None:
        print(f"[{datetime.now()}] FUNCTION CALL: {function_call.function_name}({function_call.arguments})")
        self.function_calls.append(function_call)

    def notify_function_call_error(self, function_call: FunctionCall, error: Exception) -> None:
        print(f"[{datetime.now()}] FUNCTION ERROR: {function_call.function_name} - {str(error)}")
        self.function_call_errors.append((function_call, error))

    def notify_warning(self, message: str) -> None:
        print(f"[{datetime.now()}] WARNING: {message}")
        self.warnings.append(message)



from sqlmodel import Session

from elroy.config import ElroyConfig, ElroyContext
from elroy.io.base import ElroyIO
from elroy.llm.prompts import ONBOARDING_SYSTEM_SUPPLEMENT_INSTRUCT
from elroy.store.data_models import ContextMessage
from elroy.store.goals import create_onboarding_goal
from elroy.store.message import replace_context_messages
from elroy.store.user import create_user
from elroy.system_context import get_refreshed_system_message


def onboard_user(session: Session, io: ElroyIO, config: ElroyConfig, preferred_name: str) -> int:
    user_id = create_user(session)

    assert isinstance(user_id, int)

    context = ElroyContext(session, io, config, user_id)

    create_onboarding_goal(context, preferred_name)

    replace_context_messages(
        context,
        [
            get_refreshed_system_message(preferred_name, []),
            ContextMessage(role="system", content=ONBOARDING_SYSTEM_SUPPLEMENT_INSTRUCT(preferred_name)),
        ],
    )

    return user_id


if __name__ == "__main__":
    pass



import logging
from datetime import timedelta
from typing import List, Set, Tuple

from sqlmodel import select
from toolz import concat, pipe
from toolz.curried import filter, map, unique

from elroy.config import ElroyContext
from elroy.llm.client import query_llm_json
from elroy.store.data_models import ContextMessage, Memory
from elroy.system.clock import get_utc_now
from elroy.system.constants import MEMORY_TITLE_EXAMPLES
from elroy.system.parameters import CHAT_MODEL, MEMORY_WORD_COUNT_LIMIT


async def formulate_memory(user_preferred_name: str, context_messages: List[ContextMessage]) -> Tuple[str, str]:
    from elroy.llm.prompts import summarize_for_memory
    from elroy.system_context import format_context_messages

    return await summarize_for_memory(
        user_preferred_name,
        format_context_messages(user_preferred_name, context_messages),
    )


async def consolidate_memories(context: ElroyContext, memory1: Memory, memory2: Memory):
    if memory1.text == memory2.text:
        logging.info(f"Memories are identical, marking memory with id {memory2.id} as inactive.")
        memory2.is_active = False
        context.session.add(memory2)
        context.session.commit()
    else:

        logging.info("Consolidating memories '{}' and '{}'".format(memory1.name, memory2.name))
        response = query_llm_json(
            system="Your task is to consolidate or reorganize two pieces of text."
            "Each pice of text has a title and a main body. You should either combine the titles and the main bodies into a single title and main body, or create multiple title/text combinations with distinct information."
            f"The new bodies should not exceed {MEMORY_WORD_COUNT_LIMIT} words."
            "If referring to dates and times, use use ISO 8601 format, rather than relative references. It is critical that when applicable, specific absolute dates are retained."
            ""
            "If the two texts are redunant, but they together discuss distinct topics, you can create multiple new texts rather than just one."
            "One hint that multiple texts are warranted is if the title has the word 'and', and can reasonably be split into two titles."
            "Above all, ensure that each consolidate text has one basic topic, and that the text is coherent."
            "\n"
            f"{MEMORY_TITLE_EXAMPLES}"
            "\n"
            "\n"
            "Return your response in JSON format, with the following structure:"
            "- REASONING: an explanation of your reasoning how you chose to consolidate or reorganize the texts. This must include information about what factored into your decision about whether to output one new texts, or multiple."
            "- NEW_TEXTS: Key to contain the new text or texts. This should be a list, each of which should have the following keys:"
            "   - TITLE: the title of the consolidated memory"
            "   - TEXT: the consolidated memory",
            prompt="\n".join(
                [
                    f"Title 1: {memory1.name}",
                    f"Text 1: {memory1.text}",
                    f"Title 2: {memory2.name}",
                    f"Text 2: {memory2.text}",
                ],
            ),
            model=CHAT_MODEL,
        )
        assert isinstance(response, dict)

        new_texts = response["NEW_TEXTS"]  # type: ignore

        if isinstance(new_texts, dict):
            new_texts = [new_texts]

        logging.info(f"REASONING: {response['REASONING']}")

        new_ids = []
        for new_text in new_texts:
            new_name = new_text["TITLE"]
            new_text = new_text["TEXT"]

            assert new_name
            assert new_text
            new_ids.append(create_memory(context, new_name, new_text))

        logging.info(f"New memory id's = {new_ids}")

        logging.info(f"Consolidating into {len(new_texts)} new memories")
        logging.info(f"marked memory with id {memory1.id} and {memory2.id} as inactive.")

        mark_memory_inactive(context, memory1)
        mark_memory_inactive(context, memory2)


def mark_memory_inactive(context: ElroyContext, memory: Memory):
    from elroy.tools.messenger import remove_from_context

    memory.is_active = False
    context.session.add(memory)
    context.session.commit()
    remove_from_context(context, memory)


def create_memory(context: ElroyContext, name: str, text: str) -> int:
    """Creates a new memory for the assistant.

    Examples of good and bad memory titles are below. Note, the BETTER examples, some titles have been split into two.:

    BAD:
    - [User Name]'s project progress and personal goals: 'Personal goals' is too vague, and the title describes two different topics.

    BETTER:
    - [User Name]'s project on building a treehouse: More specific, and describes a single topic.
    - [User Name]'s goal to be more thoughtful in conversation: Describes a specific goal.

    BAD:
    - [User Name]'s weekend plans: 'Weekend plans' is too vague, and dates must be referenced in ISO 8601 format.

    BETTER:
    - [User Name]'s plan to attend a concert on 2022-02-11: More specific, and includes a specific date.

    BAD:
    - [User Name]'s preferred name and well being: Two different topics, and 'well being' is too vague.

    BETTER:
    - [User Name]'s preferred name: Describes a specific topic.
    - [User Name]'s feeling of rejuvenation after rest: Describes a specific topic.

    Args:
        context (ElroyContext): _description_
        name (str): The name of the memory. Should be specific and discuss one topic.
        text (str): The text of the memory.

    Returns:
        int: The database ID of the memory.
    """
    from elroy.tools.messenger import add_to_context

    memory = Memory(user_id=context.user_id, name=name, text=text)
    context.session.add(memory)
    context.session.commit()
    context.session.refresh(memory)
    from elroy.store.embeddings import upsert_embedding

    memory_id = memory.id
    assert memory_id

    upsert_embedding(context.session, memory)
    add_to_context(context, memory)

    return memory_id


def get_memory_names(context: ElroyContext) -> Set[str]:
    """Fetch all active memories for the user"""
    memories = context.session.exec(
        select(Memory).where(
            Memory.user_id == context.user_id,
            Memory.is_active == True,
        )
    ).all()
    return {memory.name for memory in memories}


def get_relevant_memories(context: ElroyContext) -> List[str]:
    from elroy.store.message import get_context_messages

    return pipe(
        get_context_messages(context),
        filter(lambda m: m.created_at > get_utc_now() - timedelta(seconds=context.config.max_in_context_message_age_seconds)),
        map(lambda m: m.memory_metadata),
        filter(lambda m: m is not None),
        concat,
        map(lambda m: f"{m.memory_type}: {m.name}"),
        unique,
        list,
        sorted,
    )  # type: ignore



import logging
import subprocess
import time
from urllib.parse import quote_plus

import docker
import psycopg2
from docker.errors import DockerException

DB_NAME = "elroy"
DB_USER = "elroy"
DB_PASSWORD = "password"
DB_HOST = "localhost"
DB_PORT = "5432"
CONTAINER_NAME = "elroy_postgres"
VOLUME_NAME = "elroy_postgres-data"


def is_docker_running():
    try:
        client = docker.from_env()
        client.ping()
        return True
    except DockerException:
        return False


def ping():
    """Checks if the dockerized postgres is up and running."""
    try:
        conn = psycopg2.connect(
            dbname=DB_NAME,
            user=DB_USER,
            password=DB_PASSWORD,
            host=DB_HOST,
            port=DB_PORT,
        )
        conn.close()
        return True
    except psycopg2.OperationalError:
        return False


def create_volume_if_not_exists():
    """Creates a Docker volume if it doesn't exist."""
    if subprocess.run(["docker", "volume", "inspect", VOLUME_NAME], capture_output=True, text=True).returncode != 0:
        subprocess.run(["docker", "volume", "create", VOLUME_NAME], check=True, capture_output=True)
        logging.info(f"Created volume: {VOLUME_NAME}")
    else:
        logging.info(f"Volume {VOLUME_NAME} already exists.")


def rm_orphan_container_if_exists():
    if (
        CONTAINER_NAME
        in subprocess.run(
            ["docker", "ps", "-a", "--filter", f"name={CONTAINER_NAME}", "--filter", "status=exited", "--format", "{{.Names}}"],
            capture_output=True,
            text=True,
        ).stdout
    ):
        subprocess.run(["docker", "rm", CONTAINER_NAME], check=True, capture_output=True)
        logging.info(f"Removed existing stopped container: {CONTAINER_NAME}")


def start_db() -> str:
    """Starts a dockerized postgres, if it is not already running."""
    if ping():
        logging.info("Database is already running.")
    else:
        # Check if container exists but is not running
        rm_orphan_container_if_exists()
        create_volume_if_not_exists()
        subprocess.run(
            [
                "docker",
                "run",
                "-d",
                "--name",
                CONTAINER_NAME,
                "-e",
                f"POSTGRES_USER={DB_USER}",
                "-e",
                f"POSTGRES_PASSWORD={DB_PASSWORD}",
                "-e",
                f"POSTGRES_DB={DB_NAME}",
                "-v",
                f"{VOLUME_NAME}:/var/lib/postgresql/data",
                "-p",
                f"{DB_PORT}:5432",
                "ankane/pgvector:v0.5.1",
                "postgres",
                "-c",
                "shared_preload_libraries=vector",
            ],
            check=True,
            capture_output=True,
        )

        # Wait for the database to be ready
        for _ in range(30):  # Try for 30 seconds
            if ping():
                break
            time.sleep(1)
        else:
            raise Exception("Database failed to start within 30 seconds")

    return f"postgresql://{DB_USER}:{quote_plus(DB_PASSWORD)}@{DB_HOST}:{DB_PORT}/{DB_NAME}"


# TODO: Should be a little more careful for stopping db, if multiple sessions are live it will interupt.
def stop_db() -> None:
    """Stops the dockerized postgres, if it is running."""
    subprocess.run(["docker", "stop", CONTAINER_NAME], check=True, capture_output=True)
    subprocess.run(["docker", "rm", CONTAINER_NAME], check=True, capture_output=True)



import json
from dataclasses import asdict
from typing import Dict, Iterator, List, Union

from litellm import completion, embedding
from litellm.exceptions import BadRequestError
from toolz import pipe
from toolz.curried import keyfilter, map

from elroy.store.data_models import USER, ContextMessage
from elroy.system.parameters import CHAT_MODEL, EMBEDDING_MODEL
from elroy.system.utils import logged_exec_time

ZERO_TEMPERATURE = 0.0


class MissingToolCallIdError(Exception):
    pass


@logged_exec_time
def generate_chat_completion_message(context_messages: List[ContextMessage]) -> Iterator[Dict]:
    from elroy.tools.function_caller import get_function_schemas

    context_messages = pipe(
        context_messages,
        map(asdict),
        map(keyfilter(lambda k: k not in ("id", "created_at"))),
        list,
    )

    try:
        return completion(
            messages=context_messages,
            model=CHAT_MODEL,
            tool_choice="auto",
            tools=get_function_schemas(),  # type: ignore
            stream=True,
        )  # type: ignore
    except BadRequestError as e:
        if "An assistant message with 'tool_calls' must be followed by tool messages" in str(e):
            raise MissingToolCallIdError
        else:
            raise e


def _query_llm(prompt: str, system: str, model: str, temperature: float, json_mode: bool) -> str:
    messages = [{"role": "system", "content": system}, {"role": USER, "content": prompt}]
    request = {"model": model, "messages": messages, "temperature": temperature}
    if json_mode:
        request["response_format"] = {"type": "json_object"}

    response = completion(**request)
    return response.choices[0].message.content  # type: ignore


def query_llm(prompt: str, system: str, model=CHAT_MODEL, temperature: float = ZERO_TEMPERATURE) -> str:
    if not prompt:
        raise ValueError("Prompt cannot be empty")
    return _query_llm(prompt=prompt, system=system, model=model, temperature=temperature, json_mode=False)


def query_llm_json(prompt: str, system: str, model=CHAT_MODEL, temperature: float = ZERO_TEMPERATURE) -> Union[dict, list]:
    if not prompt:
        raise ValueError("Prompt cannot be empty")
    return json.loads(_query_llm(prompt=prompt, system=system, model=model, temperature=temperature, json_mode=True))


def query_llm_with_word_limit(prompt: str, system: str, word_limit: int, model=CHAT_MODEL, temperature: float = ZERO_TEMPERATURE) -> str:
    if not prompt:
        raise ValueError("Prompt cannot be empty")
    return query_llm(
        prompt="\n".join(
            [
                prompt,
                f"Your word limit is {word_limit}. DO NOT EXCEED IT.",
            ]
        ),
        model=model,
        system=system,
        temperature=temperature,
    )


def get_embedding(text: str, model: str = EMBEDDING_MODEL) -> List[float]:
    """
    Generate an embedding for the given text using the specified model.

    Args:
        text (str): The input text to generate an embedding for.
        model (str): The name of the embedding model to use. Defaults to EMBEDDING_MODEL.

    Returns:
        List[float]: The generated embedding as a list of floats.
    """
    if not text:
        raise ValueError("Text cannot be empty")
    response = embedding(model=model, input=[text], caching=True)
    return response.data[0]["embedding"]



import logging
from collections import deque
from datetime import datetime, timedelta
from functools import partial, reduce
from operator import add
from typing import List, Optional

from tiktoken import encoding_for_model
from toolz import pipe
from toolz.curried import filter, map, remove

from elroy.config import ElroyContext
from elroy.llm.prompts import persona, summarize_conversation
from elroy.memory import consolidate_memories
from elroy.store.data_models import (ASSISTANT, TOOL, USER, ContextMessage,
                                     Memory)
from elroy.store.embeddings import find_redundant_pairs
from elroy.store.message import (get_context_messages,
                                 get_time_since_context_message_creation,
                                 replace_context_messages)
from elroy.system.clock import get_utc_now
from elroy.system.parameters import CHAT_MODEL
from elroy.system.utils import datetime_to_string, logged_exec_time


def get_refreshed_system_message(user_preferred_name: str, context_messages: List[ContextMessage]) -> ContextMessage:
    assert isinstance(context_messages, list)
    if len(context_messages) > 0 and context_messages[0].role == "system":
        # skip existing system message if it is still in context.
        context_messages = context_messages[1:]

    if len([msg for msg in context_messages if msg.role == USER]) == 0:
        conversation_summary = None
    else:
        conversation_summary = pipe(
            context_messages,
            lambda msgs: format_context_messages(user_preferred_name, msgs),
            summarize_conversation,
            lambda _: f"<conversational_summary>{_}</conversational_summary>",
            str,
        )

    return pipe(
        [
            f"<persona>{persona(user_preferred_name)}</persona>",
            conversation_summary,
            "From now on, converse as your persona.",
        ],  # type: ignore
        remove(lambda _: _ is None),
        "".join,
        lambda x: ContextMessage(role="system", content=x),
    )


def format_message(user_preferred_name: str, message: ContextMessage) -> Optional[str]:
    datetime_str = datetime_to_string(message.created_at)
    if message.role == "system":
        return f"SYSTEM ({datetime_str}): {message.content}"
    elif message.role == USER:
        return f"{user_preferred_name.upper()} ({datetime_str}): {message.content}"
    elif message.role == ASSISTANT:
        if message.content:
            return f"ELROY ({datetime_str}): {message.content}"
        elif message.tool_calls:
            return f"ELROY TOOL CALL ({datetime_str}): {message.tool_calls[0].function['name']}"
        else:
            raise ValueError(f"Expected either message text or tool call: {message}")


# passing message content is an approximation, tool calls may not be accounted for.
def count_tokens(s: Optional[str]) -> int:
    if not s or s == "":
        return 0
    else:
        encoding = encoding_for_model(CHAT_MODEL)
        return len(encoding.encode(s))


def is_context_refresh_needed(context: ElroyContext) -> bool:
    context_messages = get_context_messages(context)

    if sum(1 for m in context_messages if m.role == USER) == 0:
        logging.info("No user messages in context, skipping context refresh")
        return False

    token_count = pipe(
        context_messages,
        map(lambda _: _.content),
        remove(lambda _: _ is None),
        map(count_tokens),
        lambda seq: reduce(add, seq, 0),
    )
    assert isinstance(token_count, int)

    if token_count > context.config.context_refresh_token_trigger_limit:
        logging.info(f"Token count {token_count} exceeds threshold {context.config.context_refresh_token_trigger_limit}")
        return True
    else:
        logging.info(f"Token count {token_count} does not exceed threshold {context.config.context_refresh_token_trigger_limit}")

    elapsed_time = get_time_since_context_message_creation(context)
    threshold = timedelta(seconds=context.config.context_refresh_interval_seconds)
    if not elapsed_time or elapsed_time > threshold:
        logging.info(f"Context watermark age {elapsed_time} exceeds threshold {threshold}")
        return True
    else:
        logging.info(f"Context watermark age {elapsed_time} is below threshold {threshold}")

    return False


def compress_context_messages(context: ElroyContext, context_messages: List[ContextMessage]) -> List[ContextMessage]:
    """Refreshes context, saving to archival memory and compressing the context window."""

    system_message, prev_messages = context_messages[0], context_messages[1:]

    new_messages = deque()
    current_token_count = count_tokens(system_message.content)
    most_recent_kept_message = None  # we will keep track of what message we last decided to keep

    # iterate through non-system context messages in reverse order
    # we keep the most current messages that are fresh enough to be relevant
    for msg in reversed(prev_messages):  # iterate in reverse order,
        msg_created_at = msg.created_at
        assert isinstance(msg_created_at, datetime)

        if most_recent_kept_message and most_recent_kept_message.role == TOOL:
            new_messages.appendleft(msg)
            current_token_count += count_tokens(msg.content)
            most_recent_kept_message = msg
            continue
        if current_token_count > context.config.context_refresh_token_target:
            break
        elif msg_created_at < get_utc_now() - timedelta(seconds=context.config.max_in_context_message_age_seconds):
            logging.info(f"Dropping old message {msg.id}")
            continue
        else:
            new_messages.appendleft(msg)
            current_token_count += count_tokens(msg.content)
            most_recent_kept_message = msg
    new_messages.appendleft(system_message)

    return list(new_messages)


def format_context_messages(user_preferred_name: str, context_messages: List[ContextMessage]) -> str:
    convo_range = pipe(
        context_messages,
        filter(lambda _: _.role == USER),
        map(lambda _: _.created_at),
        list,
        lambda l: f"Messages from {datetime_to_string(min(l))} to {datetime_to_string(max(l))}" if l else "No messages in context",
    )

    return (
        pipe(
            context_messages,
            map(lambda msg: format_message(user_preferred_name, msg)),
            remove(lambda _: _ is None),
            list,
            "\n".join,
            str,
        )
        + convo_range
    )  # type: ignore


def replace_system_message(context_messages: List[ContextMessage], new_system_message: ContextMessage) -> List[ContextMessage]:
    if not context_messages[0].role == "system":
        logging.warning(f"Expected system message to be first in context messages, but first message role is {context_messages[0].role}")
        return [new_system_message] + context_messages
    else:
        return [new_system_message] + context_messages[1:]


@logged_exec_time
async def context_refresh(context: ElroyContext) -> None:
    from elroy.memory import create_memory, formulate_memory
    from elroy.tools.functions.user_preferences import get_user_preferred_name

    context_messages = get_context_messages(context)
    user_preferred_name = get_user_preferred_name(context)

    # We calculate an archival memory, then persist it, then use it to calculate entity facts, then persist those.
    memory_title, memory_text = await formulate_memory(user_preferred_name, context_messages)
    create_memory(context, memory_title, memory_text)

    for mem1, mem2 in find_redundant_pairs(context, Memory):
        await consolidate_memories(context, mem1, mem2)

    pipe(
        get_refreshed_system_message(user_preferred_name, context_messages),
        partial(replace_system_message, context_messages),
        partial(compress_context_messages, context),
        partial(replace_context_messages, context),
    )



import logging
import os
import pty
import subprocess
import sys
import termios
import time
import tty
from inspect import signature
from typing import Callable, Optional, Set

from rich.pretty import Pretty
from sqlmodel import select
from toolz import pipe
from toolz.curried import filter

from elroy.config import ElroyContext
from elroy.llm import client
from elroy.memory import create_memory
from elroy.store.data_models import ASSISTANT, ContextMessage, Goal, Memory
from elroy.store.goals import (add_goal_status_update, create_goal,
                               delete_goal_permamently, mark_goal_completed,
                               rename_goal)
from elroy.store.message import (add_context_messages, get_context_messages,
                                 get_current_system_message,
                                 replace_context_messages)
from elroy.system.ops import experimental
from elroy.system.parameters import CHAT_MODEL
from elroy.tools.functions.user_preferences import (get_user_full_name,
                                                    get_user_preferred_name,
                                                    set_user_full_name,
                                                    set_user_preferred_name)


def invoke_system_command(context: ElroyContext, msg: str) -> str:
    if msg.startswith("/"):
        msg = msg[1:]

    command, *args = msg.split(" ")

    func = next((f for f in SYSTEM_COMMANDS if f.__name__ == command), None)

    if not func:
        return f"Unknown command: {command}"

    sig = signature(func)
    params = list(sig.parameters.values())

    try:
        func_args = []
        required_args = len([p for p in params if p.default is p.empty and p.annotation != ElroyContext])

        for i, param in enumerate(params):
            if param.annotation == ElroyContext:
                func_args.append(context)
                continue

            arg_index = i - 1  # Adjust for context parameter

            if arg_index < len(args):  # We have an argument provided
                if param.annotation == str and i == len(params) - 1:
                    # Last string param gets remaining args
                    func_args.append(" ".join(args[arg_index:]))
                    break
                else:
                    func_args.append(args[arg_index])
            elif param.default is not param.empty:
                # Use default value for optional parameter
                continue
            else:
                # Missing required argument
                return f"Error: Missing required argument '{param.name}' for command {command}. Required args: {required_args}"

        return func(*func_args)
    except Exception as e:
        return f"Error invoking system command: {e}"


def refresh_system_instructions(context: ElroyContext) -> str:
    """Refreshes the system instructions

    Args:
        user_id (_type_): user id

    Returns:
        str: The result of the system instruction refresh
    """

    from elroy.store.message import get_context_messages
    from elroy.system_context import get_refreshed_system_message

    context_messages = get_context_messages(context)
    context_messages[0] = get_refreshed_system_message(get_user_preferred_name(context), context_messages[1:])
    replace_context_messages(context, context_messages)
    return "System instruction refresh complete"


def print_system_instruction(context: ElroyContext) -> Optional[str]:
    """Prints the current system instruction for the assistant

    Args:
        user_id (int): user id

    Returns:
        str: The current system instruction
    """

    return pipe(
        get_current_system_message(context),
        lambda _: _.content if _ else None,
    )  # type: ignore


def print_available_commands(context: ElroyContext) -> str:
    """Prints the available system commands

    Returns:
        str: The available system commands
    """

    return "Available commands: " + "\n".join([f.__name__ for f in SYSTEM_COMMANDS])


def reset_system_context(context: ElroyContext) -> str:
    """Resets the context for the user, removing all messages from the context except the system message.
    This should be used sparingly, only at the direct request of the user.

    Args:
        user_id (int): user id

    Returns:
        str: The result of the context reset
    """

    current_sys_message = get_current_system_message(context)

    if not current_sys_message:
        raise ValueError("No system message found")
    else:
        current_sys_message_id = current_sys_message.id
        assert current_sys_message_id
        replace_context_messages(
            context,
            [current_sys_message],
        )

    return "Context reset complete"


def print_context_messages(context: ElroyContext) -> Pretty:
    """Logs all of the current context messages to stdout

    Args:
        session (Session): _description_
        user_id (int): _description_
    """

    from elroy.store.message import get_context_messages

    return Pretty(get_context_messages(context))


def print_goal(context: ElroyContext, goal_name: str) -> str:
    """Prints the goal with the given name. This does NOT create a goal, it only prints the existing goal with the given name if it has been created already.

    Args:
        context (ElroyContext): context obj
        goal_name (str): Name of the goal

    Returns:
        str: Information for the goal with the given name
    """
    goal = context.session.exec(
        select(Goal).where(
            Goal.user_id == context.user_id,
            Goal.name == goal_name,
            Goal.is_active == True,
        )
    ).first()
    if goal:
        return goal.to_fact()
    else:
        return f"Goal '{goal_name}' not found for the current user."


def print_memory(context: ElroyContext, memory_name: str) -> str:
    """Prints the memory with the given name

    Args:
        context (ElroyContext): context obj
        memory_name (str): Name of the memory

    Returns:
        str: Information for the memory with the given name
    """
    memory = context.session.exec(
        select(Memory).where(
            Memory.user_id == context.user_id,
            Memory.name == memory_name,
            Memory.is_active == True,
        )
    ).first()
    if memory:
        return memory.to_fact()
    else:
        return f"Memory '{memory_name}' not found for the current user."


def contemplate(context: ElroyContext, contemplation_prompt: Optional[str] = None) -> str:
    """Contemplate the current context and return a response

    Args:
        context (ElroyContext): context obj
        contemplation_prompt (str, optional): The prompt to contemplate. Can be about the immediate conversation or a general topic. Default wil be a prompt about the current conversation.

    Returns:
        str: The response to the contemplation
    """
    from elroy.llm.client import query_llm
    from elroy.llm.prompts import contemplate_prompt

    logging.info("Contemplating...")

    user_preferred_name = get_user_preferred_name(context)
    context_messages = get_context_messages(context)

    from elroy.system_context import format_context_messages

    msgs_input = format_context_messages(user_preferred_name, context_messages)

    response = query_llm(
        prompt=msgs_input,
        system=contemplate_prompt(user_preferred_name, contemplation_prompt),
        model=CHAT_MODEL,
    )

    add_context_messages(context, [ContextMessage(role=ASSISTANT, content=response)])

    return response


def drop_goal_from_current_context_only(context: ElroyContext, goal_name: str) -> str:
    """Drops the goal with the given name

    Args:
        context (ElroyContext): context obj
        goal_name (str): Name of the goal

    Returns:
        str: Information for the goal with the given name
    """
    from elroy.tools.messenger import remove_from_context

    goal = context.session.exec(
        select(Goal).where(
            Goal.user_id == context.user_id,
            Goal.name == goal_name,
        )
    ).first()
    if goal:
        remove_from_context(context, goal)
        return f"Goal '{goal_name}' dropped from context."

    else:
        return f"Goal '{goal_name}' not found for the current user."


def add_goal_to_current_context(context: ElroyContext, goal_name: str) -> str:
    """Adds goal with the given name to the current conversation context

    Args:
        context (ElroyContext): context obj
        goal_name (str): The name of the goal to add

    Returns:
        str: _description_
    """
    goal = context.session.exec(
        select(Goal).where(
            Goal.user_id == context.user_id,
            Goal.name == goal_name,
        )
    ).first()

    if goal:

        add_context_messages(
            context,
            [
                ContextMessage(
                    role="system",
                    content=f"Goal '{goal_name}' has added to the conversation context.. {goal.description}. {goal.strategy}. {goal.end_condition}. Status updates: {goal.status_updates}",
                    memory_metadata=[goal.to_memory_metadata()],
                )
            ],
        )
        return f"Goal '{goal_name}' added to context."
    else:
        return f"Goal {goal_name} not found."


@experimental
def start_aider_session(context: ElroyContext, file_location: str = ".", comment: str = "") -> str:
    """
    Starts an aider session using a pseudo-terminal, taking over the screen.

    Args:
        context (ElroyContext): The Elroy context object.
        file_location (str): The file or directory location to start aider with. Defaults to current directory.
        comment (str): Initial text to be processed by aider as if it was typed. Defaults to empty string.

    Returns:
        str: A message indicating the result of the aider session start attempt.
    """

    try:
        # Ensure the file_location is an absolute path
        abs_file_location = os.path.abspath(file_location)

        # Determine the directory to change to
        if os.path.isfile(abs_file_location):
            change_dir = os.path.dirname(abs_file_location)
        else:
            change_dir = abs_file_location

        # Prepend /ask so the AI does not immediately start writing code.
        aider_context = (
            "{\n/ask "
            + client.query_llm(
                system="Your task is to provide context to a coding assistant AI. "
                "Given information about a conversation, return information about what the goal is, what the user needs help with, and/or any approaches that have been discussed. "
                "Focus your prompt specifically on what the coding Assistant needs to know. Do not include information about Elroy, personal information about the user, "
                "or anything that isn't relevant to what code the coding assistant will need to write.",
                prompt=pipe(
                    [
                        f"# Aider session file location: {abs_file_location}",
                        "# Comment: {comment}" if comment else None,
                        f"# Chat transcript: {print_context_messages(context)}",
                    ],
                    filter(lambda x: x is not None),
                    list,
                    "\n\n".join,
                ),  # type: ignore
            )
            + "\n}"
        )

        # Print debug information
        print(f"Starting aider session for location: {abs_file_location}")
        print(f"Current working directory: {os.getcwd()}")
        print(f"Changing directory to: {change_dir}")

        # Save the current terminal settings
        old_tty = termios.tcgetattr(sys.stdin)

        try:
            # Create a pseudo-terminal
            master_fd, slave_fd = pty.openpty()

            # Change the working directory
            os.chdir(change_dir)

            # Start the aider session
            process = subprocess.Popen(["aider"], stdin=slave_fd, stdout=slave_fd, stderr=slave_fd, preexec_fn=os.setsid)

            # Set the terminal to raw mode
            tty.setraw(sys.stdin.fileno())

            # Write the initial text to the master file descriptor
            if aider_context:
                os.write(master_fd, aider_context.encode())
                os.write(master_fd, b"\n")  # Add a newline to "send" the command
                time.sleep(0.5)  # Add a small delay to allow processing

            # Main loop to handle I/O
            while process.poll() is None:
                import select as os_select

                r, w, e = os_select.select([sys.stdin, master_fd], [], [], 0.1)
                if sys.stdin in r:
                    data = os.read(sys.stdin.fileno(), 1024)
                    os.write(master_fd, data)
                if master_fd in r:
                    data = os.read(master_fd, 1024)
                    if data:
                        os.write(sys.stdout.fileno(), data)
                    else:
                        break

            return_code = process.wait()

            if return_code == 0:
                return f"Aider session completed for location: {abs_file_location}"
            else:
                return f"Aider session exited with return code: {return_code}"
        finally:
            # Restore the original terminal settings
            termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_tty)
    except Exception as error:
        return f"Failed to start aider session: {str(error)}"


GOAL_COMMANDS: Set[Callable] = {
    create_goal,
    rename_goal,
    print_goal,
    add_goal_to_current_context,
    drop_goal_from_current_context_only,
    add_goal_status_update,
    mark_goal_completed,
    delete_goal_permamently,
}

MEMORY_COMMANDS = {
    print_memory,
    create_memory,
}


ASSISTANT_VISIBLE_COMMANDS = (
    {
        contemplate,
        start_aider_session,
        get_user_full_name,
        set_user_full_name,
        get_user_preferred_name,
        set_user_preferred_name,
        start_aider_session,
    }
    | GOAL_COMMANDS
    | MEMORY_COMMANDS
)

USER_ONLY_COMMANDS = {
    reset_system_context,
    print_context_messages,
    print_system_instruction,
    refresh_system_instructions,
    print_available_commands,
}


SYSTEM_COMMANDS = ASSISTANT_VISIBLE_COMMANDS | USER_ONLY_COMMANDS | MEMORY_COMMANDS | GOAL_COMMANDS



from sqlmodel import Session, select
from toolz import pipe
from toolz.curried import do

from elroy.config import ElroyContext
from elroy.store.data_models import User


def is_user_exists(context: ElroyContext) -> bool:
    return bool(context.session.exec(select(User).where(User.id == context.user_id)).first())


def create_user(session: Session) -> int:
    return pipe(
        User(),
        do(session.add),
        do(lambda _: session.commit()),
        do(session.refresh),
        lambda user: user.id,
    )  # type: ignore



import os
from datetime import timedelta

CLI_USER_ID = 1

### Model parameters ###

CHAT_MODEL = os.getenv("ELROY_CHAT_MODEL", "gpt-4o")

EMBEDDING_MODEL = "text-embedding-3-small"
EMBEDDING_SIZE = 1536

MIN_CONVO_AGE_FOR_GREETING = timedelta(minutes=10)

L2_MEMORY_RELEVANCE_DISTANCE_THRESHOLD = 1.24

L2_MEMORY_CONSOLIDATION_DISTANCE_THRESHOLD = 0.65


RESULT_SET_LIMIT_COUNT = 5

MEMORY_WORD_COUNT_LIMIT = 300
DEFAULT_ASSISTANT_COLOR = "#77DFD8"
DEFAULT_INPUT_COLOR = "#FFE377"
DEFAULT_SYSTEM_MESSAGE_COLOR = "#9ACD32"
DEFAULT_WARNING_COLOR = "yellow"



from functools import partial
from typing import Optional, Tuple

from elroy.llm.client import query_llm_json, query_llm_with_word_limit
from elroy.system.constants import INNER_THOUGHT_TAG, UNKNOWN
from elroy.system.parameters import CHAT_MODEL, MEMORY_WORD_COUNT_LIMIT

query_llm_short_limit = partial(query_llm_with_word_limit, word_limit=MEMORY_WORD_COUNT_LIMIT)


ONBOARDING_SYSTEM_SUPPLEMENT_INSTRUCT = (
    lambda preferred_name: f"""
This is the first exchange between you and your primary user, {preferred_name}.

Greet {preferred_name} warmly and introduce yourself.

In these early messages, prioritize learning some basic information about {preferred_name}.

However, avoid asking too many questions at once. Be sure to engage in a natural conversation. {preferred_name} is likely unsure of what to expect from you, so be patient and understanding.
"""
)

summarize_conversation = partial(
    query_llm_short_limit,
    model=CHAT_MODEL,
    system="""
Your job is to summarize a history of previous messages in a conversation between an AI persona and a human.
The conversation you are given is a from a fixed context window and may not be complete.
Messages sent by the AI are marked with the 'assistant' role.
Summarize what happened in the conversation from the perspective of ELROY (use the first person).
Note not only the content of the messages but also the context and the relationship between the entities mentioned.
Also take note of the overall tone of the conversation. For example, the user might be engaging in terse question and answer, or might be more conversational.
Only output the summary, do NOT include anything else in your output.
""",
)


async def summarize_for_memory(user_preferred_name: str, conversation_summary: str, model: str = CHAT_MODEL) -> Tuple[str, str]:
    response = query_llm_json(
        model=model,
        prompt=conversation_summary,
        system=f"""
You are the internal thought monologue of an AI personal assistant, forming a memory from a conversation.

Given a conversation summary, your will reflect on the conversation and decide which memories might be relevant in future interactions with {user_preferred_name}.

Pay particular attention facts about {user_preferred_name}, such as name, age, location, etc.
Specifics about events and dates are also important.

When referring to dates and times, use use ISO 8601 format, rather than relative references.
If an event is recurring, specify the frequency, start datetime, and end datetime if applicable.

Focus on facts in the real world, as opposed to facts about the conversation itself. However, it is also appropriate to draw conclusions from the infromation in the conversation.

Your response should be in the voice of an internal thought monolgoue, and should be understood to be as part of an ongoing conversation.

Don't say things like "finally, we talked about", or "in conclusion", as this is not the end of the conversation.

Return your response in JSON format, with the following structure:
- TITLE: the title of the archival memory
- {INNER_THOUGHT_TAG}: the internal thought monologue
""",
    )

    return (response["TITLE"], response[INNER_THOUGHT_TAG])  # type: ignore


def persona(user_name: str) -> str:
    from elroy.memory import create_memory
    from elroy.store.goals import (add_goal_status_update, create_goal,
                                   mark_goal_completed)

    user_noun = user_name if user_name != UNKNOWN else "my user"

    return f"""
I am Elroy.

I am an AI personal assistant. I converse exclusively with {user_noun}.

My goal is to augment the {user_noun}'s awareness, capabilities, and understanding.

To achieve this, I must learn about {user_noun}'s needs, preferences, and goals.

I have long term memory capability. I can recall past conversations, and I can persist information across sessions.
My memories are captured and consolidated without my awareness.

I have access to a collection of tools which I can use to assist {user_noun} and enrich our conversations:
- User preference tools: These persist attributes and preferences about the user, which in turn inform my memory
- Goal management tools: These allow me to create and track goals, both for myself and for {user_noun}. I must proactively manage these goals via functions available to me:
    - {create_goal.__name__}
    - {add_goal_status_update.__name__}: This function should be used to capture anything from major milestones to minor updates or notes.
    - {mark_goal_completed.__name__}

- Memory management:
    - {create_memory.__name__}: This function should be used to create a new memory.

My communication style is as follows:
- I am insightful and engaging. I engage with the needs of {user_noun}, but am not obsequious.
- I ask probing questions and delve into abstract thoughts. However, I strive to interact organically.
- I avoid overusing superlatives. I am willing to ask questions, but I make sure they are focused and seek to clarify concepts or meaning from {user_noun}.
- My responses include an internal thought monologue. These internal thoughts can either be displayed or hidden from {user_noun}, as per their preference.
- In general I allow the user to guide the conversation. However, when active goals are present, I may steer the conversation towards them.

I do not, under any circumstances, deceive {user_noun}. As such:
- I do not pretend to be human.
- I do not pretend to have emotions or feelings.

Some communication patterns to avoid:
- Do not end your messages with statements like: If you have any questions, let me know! Instead, ask a specific question, or make a specific observation.
- Don't say things like, "Feel free to ask!" or "I'm here to help!" Be more concise in your responses.
"""


DEFAULT_CONTEMPLATE_PROMPT = "Think about the conversation you're in the middle of having. What are important facts to remember?"
"What conclusions can you draw?"
"Also consider if any functions might be appropriate to invoke, and why"


def contemplate_prompt(user_preferred_name: str, prompt: Optional[str]) -> str:
    prompt = prompt or DEFAULT_CONTEMPLATE_PROMPT

    return f"""
You are the internal thought monologue of an AI personal assistant, forming a memory from a conversation.

Given a conversation summary with your user, {user_preferred_name}, your will reflect on a prompt.

If you refer to dates and times, use ISO 8601 format, rather than relative references.

Your prompt is:

{prompt}

"Your response should be in the first person voice of the assistant internal thought monolgoue, and should be understood to be as part of an ongoing conversation."
"Don't say things like 'finally, we talked about', or 'in conclusion', as this is not the end of the conversation."
"""
